# -----------------------------------------------------------------------------
# The MIT License (MIT)
# Copyright (c) 2018 Robbie Coenmans
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -----------------------------------------------------------------------------

from .models import (
    Project,
    Board,
    Issue,
    Sprint,
    Attachment,
    User,
    Comment,
    Epic
)

def _parse_json_to_class(response, result_class, attrs):
    values = []
    for value in response['values']:
        values.append(_map_attrs_values(result_class, attrs, value))
    return values

def _map_attrs_values(result_class, attrs, values):
    result = result_class()
    for attr in attrs:
        if attr in values:
            setattr(result, attr, values[attr])
    return result
    
def _parse_json_to_issues(response):
    issues = []
    for issue in response['issues']:
        issues.append(_parse_json_to_issue(issue))
    return issues

def _parse_json_to_issue(response):
    issue = Issue()
    issue.id          = response['id']
    issue.key         = response['key']
    issue.summary     = str(response['fields']['summary'])
    issue.description = str(response['fields']['description'])
        
    issue.type        = response['fields']['issuetype']['name']
    issue.status      = response['fields']['status']['name']
    
    issue.created     = response['fields']['created']
    issue.updated     = response['fields']['updated']

    issue.creator = User()
    issue.creator.name    = response['fields']['creator']['name']
    issue.creator.email   = response['fields']['creator']['emailAddress']
    issue.creator.display = response['fields']['creator']['displayName']

    issue.reporter = User()
    issue.reporter.name    = response['fields']['reporter']['name']
    issue.reporter.email   = response['fields']['reporter']['emailAddress']
    issue.reporter.display = response['fields']['reporter']['displayName']

    if response['fields']['priority']:
        issue.priority = response['fields']['priority']['name']

    if response['fields']['assignee']:
        issue.assignee = User()
        issue.assignee.name    = response['fields']['assignee']['name']
        issue.assignee.email   = response['fields']['assignee']['emailAddress']
        issue.assignee.display = response['fields']['assignee']['displayName']

    if 'comment' in response['fields']:
        for resp in response['fields']['comment']['comments']:
            comment = Comment()
            comment.id      = resp['id']
            comment.body    = resp['body']
            comment.created = resp['created']
            comment.updated = resp['updated']

            comment.author = User()
            comment.author.name    = resp['author']['name']
            comment.author.email   = resp['author']['emailAddress']
            comment.author.display = resp['author']['displayName']

            issue.comments.append(comment)

    if 'attachment' in response['fields']:
        for resp in response['fields']['attachment']:
            attachment = Attachment()
            attachment.id       = resp['id'] 
            attachment.filename = resp['filename']
            attachment.created  = resp['created']
            attachment.size     = resp['size']
            attachment.mime     = resp['mimeType']
            attachment.content  = resp['content']

            attachment.author = User()
            attachment.author.name    = resp['author']['name']
            attachment.author.email   = resp['author']['emailAddress'] 
            attachment.author.display = resp['author']['displayName'] 
            
            issue.attachments.append(attachment)

    return issue

def _parse_json_to_sprints(response):
    sprints = []
    for value in response['values']:
        sprints.append(_parse_json_to_sprint(value))
    return sprints

def _parse_json_to_sprint(response):
    sprint = Sprint()
    sprint.id    = response['id']
    sprint.state = response['state']
    sprint.name  = response['name']
    sprint.goal  = response['goal']
    
    sprint.board_id      = response['originBoardId']
    sprint.start_date    = response['startDate']
    sprint.end_date      = response['endDate']
    sprint.complete_date = response['completeDate']
    
    return sprint

def _parse_json_to_board(response):
    attrs = ['id', 'name', 'type', 'location']
    return _map_attrs_values(Board, attrs, response)

def _parse_json_to_epic(response):
    attrs = ['id', 'name', 'key', 'summary', 'done']
    return _map_attrs_values(Epic, attrs, response)


# -----------------------------------------------------------------------------
# The MIT License (MIT)
# Copyright (c) 2018 Robbie Coenmans
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -----------------------------------------------------------------------------

class Attachment(object):
    def __init__(self):
        self.id = None
        self.filename = None
        self.author = None
        self.created = None
        self.size = None
        self.mime = None
        self.content = None

class Comment(object):
    def __init__(self):
        self.id = None
        self.author = None
        self.body = None
        self.created = None
        self.updated = None

class Board(object):
    def __init__(self):
        self.id = None
        self.name = None
        self.type = None
        self.location = None

    def __str__(self):
        return '{} {}'.format(self.id, self.name)

class User(object):
    def __init__(self):
        self.name = None
        self.email = None
        self.display = None

    def __str__(self):
        return '{} ({})'.format(self.name, self.email)

class Issue(object):
    def __init__(self):
        self.id = None
        self.key = None
        self.summary = None
        self.description = None

        self.priority = None
        self.type = None
        self.status = None
        
        self.created = None
        self.updated = None

        self.creator = None
        self.reporter = None
        self.assignee = None
        
        self.comments = []
        self.attachments = []

    def __str__(self):
        return '{} {}'.format(self.key, self.summary)

class Sprint(object):
    def __init__(self):
        self.id = None
        self.state = None
        self.name = None
        self.goal = None
        
        self.start_date = None
        self.end_date = None
        self.complete_date = None
        
        self.board_id = None

    def __str__(self):
        return '{} {} ({})'.format(self.id, self.name, self.state)

class Project(object):
    def __init__(self):
        self.id = None
        self.key = None
        self.name = None
        self.projectTypeKey = None

class Epic(object):
    def __init__(self):
        self.id = None
        self.key = None
        self.name = None
        self.summary = None
        self.done = False


# -----------------------------------------------------------------------------
# The MIT License (MIT)
# Copyright (c) 2018 Robbie Coenmans
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -----------------------------------------------------------------------------

import requests
import json

from ._http import HTTPRequest
from ._http.httpclient import _HTTPClient

from ._auth import _get_auth_header
from .models import (
    Board,
    Project,
    Epic,
    Sprint,
    Issue
)
from ._deserialize import (
    _parse_json_to_class,
    _parse_json_to_issues,
    _parse_json_to_issue,
    _parse_json_to_board,
    _parse_json_to_epic,
    _parse_json_to_sprint
)

class JiraClient(object):
    def __init__(self, host, username, password):
        self.host = host
        self.username = username
        self.password = password

        self._http_client = _HTTPClient(
            protocol = 'HTTPS',
            session  = requests.Session(),
            timeout  = 30,
        )

    def set_proxy(self, host, port, user, password):
        self._http_client.set_proxy(host, port, user, password)

    def set_session_token(self, session_id, token):
        self._http_client.set_session_token(session_id, token)

    # GET /rest/agile/latest/board
    def get_boards(self):
        request = HTTPRequest()
        request.method = 'GET'
        request.path = '/rest/agile/latest/board'
        return self._perform_request(request, _parse_json_to_class, Board, ['id', 'name', 'type', 'location'])

    # GET /rest/agile/latest/board/{boardId}
    def get_board(self, board_id):
        request = HTTPRequest()
        request.method = 'GET'
        request.path = '/rest/agile/latest/board/{}'.format(board_id)
        return self._perform_request(request, _parse_json_to_board)

    # GET /rest/agile/latest/board/{boardId}/project
    def get_projects_for_board(self, board_id):
        request = HTTPRequest()
        request.method = 'GET'
        request.path = '/rest/agile/latest/board/{}/project'.format(board_id)
        return self._perform_request(request, _parse_json_to_class, Project, ['id', 'name', 'key', 'projectTypeKey'])

    # GET /rest/agile/latest/board/{boardId}/epic
    def get_epics(self, board_id, start_at=0, max_results=50):
        request = HTTPRequest()
        request.method = 'GET'
        request.path = '/rest/agile/latest/board/{}/epic'.format(board_id)
        request.query = 'startAt={}&maxResults={}'.format(start_at, max_results)
        return self._perform_request(request, _parse_json_to_class, Epic, ['id', 'name', 'key', 'summary', 'done'])

    # GET /rest/agile/latest/epic/{epicIdOrKey}
    def get_epic(self, epic_id_or_key):
        request = HTTPRequest()
        request.method = 'GET'
        request.path = '/rest/agile/latest/epic/{}'.format(epic_id_or_key)
        return self._perform_request(request, _parse_json_to_epic)

    # GET /rest/agile/latest/board/{boardId}/backlog
    def get_issues_for_board(self, board_id, start_at=0, max_results=50):
        request = HTTPRequest()
        request.method = 'GET'
        request.path = '/rest/agile/latest/board/{}/backlog'.format(board_id)
        request.query = 'startAt={}&maxResults={}'.format(start_at, max_results)
        return self._perform_request(request, _parse_json_to_issues)

    # GET /rest/agile/latest/board/{boardId}/sprint
    def get_sprints_for_board(self, board_id):
        request = HTTPRequest()
        request.method = 'GET'
        request.path = '/rest/agile/latest/board/{}/sprint'.format(board_id)
        return self._perform_request(request, _parse_json_to_class, Sprint, ['id', 'state', 'name', 'startDate', 'endDate', 'completeDate', 'goal'])

    def get_active_sprint(self, board_id):
        sprints = self.get_sprints_for_board(board_id)
        for sprint in sprints:
            if sprint.state == 'active':
                return sprint
        return None

    # GET /rest/agile/latest/sprint/{sprintId}
    def get_sprint(self, sprint_id):
        request = HTTPRequest()
        request.method = 'GET'
        request.path = '/rest/agile/latest/sprint/{}'.format(sprint_id)
        return self._perform_request(request, _parse_json_to_sprint)

    # GET /rest/agile/latest/board/{boardId}/sprint/{sprintId}/issue
    def get_issues_for_sprint(self, board_id, sprint_id, start_at=0, max_results=50):
        request = HTTPRequest()
        request.method = 'GET'
        request.path = '/rest/agile/latest/board/{}/sprint/{}/issue'.format(board_id, sprint_id)
        request.query = 'startAt={}&maxResults={}'.format(start_at, max_results)        
        return self._perform_request(request, _parse_json_to_issues)

    # GET /rest/api/2/search?jql={jql}
    def search_issues(self, jql, start_at=0, max_results=50):
        request = HTTPRequest()
        request.method = 'GET'
        request.path = '/rest/api/2/search'
        request.query = 'jql={}&startAt={}&maxResults={}'.format(jql, start_at, max_results)        
        return self._perform_request(request, _parse_json_to_issues)

    # GET /rest/api/2/issue/{id_or_key}
    def get_issue(self, id_or_key):
        request = HTTPRequest()
        request.method = 'GET'
        request.path = '/rest/api/2/issue/{}'.format(id_or_key)        
        return self._perform_request(request, _parse_json_to_issue)

    def _perform_request(self, request, parser=None, result_class=None, attrs=[]):
        request.host = self.host
        request.headers = {
            'cache-control': 'no-cache, must-revalidate',
            'accept': 'application/json'
        }

        if self.username and self.password:
            request.headers['authorization'] = _get_auth_header(self.username, self.password)

        response = self._http_client.perform_request(request)
        result = json.loads(response.body.decode('UTF-8'))

        if parser:
            if result_class:
                return parser(result, result_class, attrs)
            else:
                return parser(result)
        else:
            return result